/**
 * @file		DatFileEntry.hpp
 * @date		04/02/2023
 * @author		JackCarterSmith
 * @copyright	GPL-v3.0
 * @brief		Data file entry descriptor objects helper class.
 *
 * DATA and BUNDLE file store in different location file's header and
 * file's data. This helper class it's used to provide a more convenient
 * access to the corresponding file inside memory (after Erso module's
 * initialization).
 *
 */

#include <FileHandler/Generic.hpp>
#include <vector>
#include <string>


#ifndef DATFILEENTRY_HPP_
#define DATFILEENTRY_HPP_


namespace RDI {
namespace DatFile {

typedef union u_file_flags {
	struct {
		// LSB
		unsigned short unknown0:1;
		unsigned short isFile:1;
		unsigned short unknown1:5;
		unsigned short isDirectory:1;
		unsigned short unknown2:8;
		// MSB
	};
	unsigned short raw;
} DAT_FILE_FLAGS;



class DatFileEntry {
public:
	virtual ~DatFileEntry() = 0;
	virtual bool isDirectory() = 0;
	virtual bool isRootDirectory() = 0;
	virtual unsigned int getSize() = 0;

	std::string getName() { return name; }
	std::string getPath() { return virtualAbsPath; }

protected:
	DatFileEntry( std::string name, std::string path ) : name(name), virtualAbsPath(path) {
		this->fFlags.raw = 0;
	}
	DatFileEntry( std::string name, std::string path, DAT_FILE_FLAGS fFlags )
			: name(name), virtualAbsPath(path), fFlags(fFlags) {}

	std::string name;
	std::string virtualAbsPath;
	DAT_FILE_FLAGS fFlags;
};

inline DatFileEntry::~DatFileEntry() {}

class DatFileEntryFile : public DatFileEntry {
public:
	DatFileEntryFile( std::string name, std::string path, DAT_FILE_FLAGS fFlags, unsigned int size, MEMFILE fPtr );
	~DatFileEntryFile();

	MEMFILE getDatas() { return fileMemPtr; }

	bool isDirectory() { return false; }
	bool isRootDirectory() { return false; }
	unsigned int getSize() { return size; }		// In bytes

private:
	unsigned int size;
	MEMFILE fileMemPtr;
};


class DatFileEntryDirectory : public DatFileEntry {
public:
	DatFileEntryDirectory( std::string name, std::string path );
	DatFileEntryDirectory( std::string name, std::string path, DAT_FILE_FLAGS fFlags );
	DatFileEntryDirectory( std::string name, std::string path, DAT_FILE_FLAGS fFlags, DatFileEntry* hFileEntry );
	~DatFileEntryDirectory();

	void ClearLinkedFiles();
	void AddEntry( DatFileEntry* hFileEntry );
	std::vector<DatFileEntry*> getFiles() { return *vSubFiles; }

	bool isDirectory() { return true; }
	bool isRootDirectory() { return rootDir; }
	unsigned int getSize() { return vSubFiles->size(); }	// In number of elements

private:
	bool rootDir;
	std::vector<DatFileEntry*> *vSubFiles;
};

}
}

#endif /* DATFILEENTRY_HPP_ */
