/**
 * @file		Erso.cpp
 * @date		04/02/2023
 * @author		JackCarterSmith
 * @copyright	GPL-v3.0
 * @brief		Rogue DATA/BUNDLE file parser and interface.
 *
 */

#include <stdlib.h>
#include <iostream>
#include <fstream>
#include <vector>
#include <boost/filesystem.hpp>
#include <RDI.hpp>	//TODO: Should be removed with RDI_RESULT
#include "files_mms/data_struct.h"
#include "DatFileEntry.hpp"
#include "Erso.hpp"
#include <FileHandler/Generic.hpp>

using std::string;
using std::ios;
using std::vector;
using boost::filesystem::path;
using boost::filesystem::file_size;

using namespace RDI::DatFile;


namespace RDI {

Erso::Erso( string fPath ) : workingDir(fPath)
{
	Process();
}

Erso::Erso( string fPath, RDI_RESULT* outResult ) :
		workingDir(fPath)
{
	if (outResult) *outResult = Process();
	else Process();
}

Erso::~Erso()
{
	DisposeFile();
}


string Erso::getDataSectionName( unsigned char id ) const { return pDatSection->at(id).name; }
unsigned int Erso::getDataSectionOffset( unsigned char id ) const { return pDatSection->at(id).offset; }

/**
 * Change the current working directory.
 * This action will reconstruct files tree, all previous file entries became invalid.
 *
 * @param newPath	New working directory path
 *
 * @return Error status.
 */
RDI_RESULT Erso::setWorkingDirectory( std::string newPath ) {
	RDI_RESULT errCode = RDI_OK;

	DisposeFile();

	this->workingDir = newPath;
	Process();

	return errCode;
}


/* ------------------------------------------------------------------------------------------------------------- */
RDI_RESULT Erso::Process() {
	RDI_RESULT errCode = RDI_OK;

	// Process header file and dump data file in memory.
	errCode = DataFilesDumping();
	if (errCode == RDI_ERROR_FILESYSTEM) {
		std::cout << "Data files (DATA.DAT/DATA.HDR) not found! Interrupt." << std::endl;
		return errCode;
	} else if (errCode == RDI_ERROR_MEMORY) {
		std::cout << "Memory allocation or file access failed! Interrupt." << std::endl;
		return errCode;
	}

	// Process data file tree and filter file subtype into list.
	CreateDataFilesTree();

	return errCode;
}

/**
 * Helper function to search for legacy DATA.DAT/HDR and dump it in memory.
 *
 * @return Error status.
 */
RDI_RESULT Erso::DataFilesDumping() {
	unsigned int i;
	path fp(workingDir);
	MEMFILE pTmpFile;

	// If input as a directory, assume it's contain DATA.DAT/HDR.
	if (boost::filesystem::is_directory(fp)) fp.append("DATA.HDR");

	// If file specified in input is the "datas" file, process the header (HDR) before.
	if (fp.extension() == ".DAT") fp.replace_extension(".HDR");

	// Open and parse data header file.
	//if (boost::filesystem::exists(fp)) return RDI_ERROR_FILESYSTEM;
	pTmpFile = CreateMemFile(fp);
	if (!pTmpFile) return RDI_ERROR_MEMORY;

	cDatSection = file_size(fp) / sizeof(T_HDR_ENTRY);
	pDatSection = new vector<struct dataSection>(cDatSection);
	for ( i = 0; i < pDatSection->size(); i++ ) {
		// Store header infos into structure to make easier access later.
		pDatSection->at(i).name.append(((T_HDR_ENTRY*)(pTmpFile + i * sizeof(T_HDR_ENTRY)))->section_name);
		pDatSection->at(i).offset = ((T_HDR_ENTRY*)(pTmpFile + i * sizeof(T_HDR_ENTRY)))->section_offset;
	}

	free(pTmpFile);

	// Dump data file and store it's pointer into class member.
	fp.replace_extension(".DAT");
	//if (boost::filesystem::exists(fp)) return RDI_ERROR_FILESYSTEM;
	rDatPtr = CreateMemFile(fp);
	if (!rDatPtr) return RDI_ERROR_MEMORY;

	return RDI_OK;
}

/**
 * Dump file into to newly allocated memory.
 * @param filePath File path.
 *
 * @return Start memory pointer of the file.
 */
MEMFILE Erso::CreateMemFile( path filePath ) {
	const unsigned int size = file_size(filePath);
	std::fstream rdf;
	MEMFILE fPtr = NULL;

	rdf.open(filePath.string(), ios::in | ios::binary);
	if (rdf.is_open()) {
		fPtr = (MEMFILE)malloc(size);
		rdf.read(fPtr, size);
		rdf.close();
	}

	return fPtr;
}

/**
 * Helper function to process files content for further use in this library.
 *
 * @return Error status.
 */
RDI_RESULT Erso::CreateDataFilesTree() {
	unsigned int curEntriesCount = 0, i;
	DatFileEntryDirectory* curDir = nullptr;
	MEMFILE pFileDescriptors = nullptr;

	// Create new root files tree. One per data section.
	pFTRoot = new std::vector<DatFileEntryDirectory*>;
	pFTRoot->clear();
	for ( i = 0; i < cDatSection; i++ ) {
		curDir = new DatFileEntryDirectory(pDatSection->at(i).name, pDatSection->at(i).name);
		pFTRoot->push_back(curDir);

		// Recalculate files descriptor offset for current section.
		pFileDescriptors = rDatPtr + pDatSection->at(i).offset +
				((T_DAT_SECTION*)(rDatPtr + pDatSection->at(i).offset))->file_headers_offset;
		curEntriesCount = (((T_DAT_SECTION*)(rDatPtr + pDatSection->at(i).offset))->file_headers_size) / 32;

		// Process arborescente in recursive method.
		ProcessTreeDirectoryContents(curDir, pFileDescriptors, curEntriesCount, rDatPtr + pDatSection->at(i).offset);
	}

	return RDI_OK;
}

/**
 * Recursive function to process all sub-directory from a specific DatFileEntryDirectory root.
 *
 * @param pCurDir		Root/current DatFileEntryDirectory from which the function begin to dig into
 * @param pDesc			MEMFILE pointer to the first entry of directory to be parsed
 * @param count			Number of elements in the pCurDir directory
 * @param pSectionStart MEMFILE pointer to the beginning of parsed file
 *
 * @return Error status.
 */
RDI_RESULT Erso::ProcessTreeDirectoryContents( DatFileEntryDirectory* pCurDir, MEMFILE pDesc, const unsigned int count, MEMFILE pSectionStart ) {
	unsigned int i, newDRECnt;
	DAT_FILE_FLAGS curEntryFlags = {0};

	DatFileEntryFile* newFile = nullptr;
	DatFileEntryDirectory* newDir = nullptr;
	std::string tmpStr;

	if ( count == 0 ) return RDI_ERROR_PROCESS;

	for ( i = 0; i < count; i++ ) {
		curEntryFlags.raw = ((T_FILE_HEADER*)(pDesc + i * sizeof(T_FILE_HEADER)))->flags;

		tmpStr.clear();
		tmpStr.append(((T_FILE_HEADER*)(pDesc + i * sizeof(T_FILE_HEADER)))->name);

		// Test for file or directory
		if (!curEntryFlags.isDirectory) {
			newFile = new DatFileEntryFile(tmpStr, pCurDir->getPath() + "/" + tmpStr, curEntryFlags,
					((T_FILE_HEADER*)(pDesc + i * sizeof(T_FILE_HEADER)))->datas_size,
					pSectionStart + ((T_FILE_HEADER*)(pDesc + i * sizeof(T_FILE_HEADER)))->datas_offset);
			pCurDir->AddEntry(newFile);
		} else {
			newDir = new DatFileEntryDirectory(tmpStr, pCurDir->getPath() + "/" + tmpStr, curEntryFlags);
			pCurDir->AddEntry(newDir);

			// Keep entries count in new directory to jump over after processing it.
			newDRECnt = (((T_FILE_HEADER*)(pDesc + i * sizeof(T_FILE_HEADER)))->dir_entries_size - sizeof(T_FILE_HEADER)) / sizeof(T_FILE_HEADER);
			ProcessTreeDirectoryContents(newDir, pDesc + (i + 1) * sizeof(T_FILE_HEADER), newDRECnt, pSectionStart);
			if (newDRECnt <= count) i += newDRECnt;
		}
	}

	return RDI_OK;
}

/**
 * Free all allocated resources for files tree and MemFile
 */
void Erso::DisposeFile() {
	for ( DatFileEntryDirectory* de : *pFTRoot )
		delete de;
	delete pFTRoot;
	free(rDatPtr);
	delete pDatSection;
}


/* ------------------------------------------------------------------------------------------------------------- */
/**
 * Get DatFileEntry reference from virtual path
 *
 * @param virtualDirPath	Path to the virtual resource
 *
 * @return	Reference to the resource from specified input path
 */
DatFileEntry *Erso::getElement( path virtualDirPath ) {
	bool skip = false;
	path newPath;

	for ( DatFileEntry* de : *pFTRoot ) {
		if (virtualDirPath.size() == 1 || virtualDirPath.begin() == --virtualDirPath.end()) {
			if (de->getName() == virtualDirPath.filename().string())
				return de;
		} else {
			if (de->getName() == virtualDirPath.begin()->string()) {
				for (auto& sp : virtualDirPath) {
									if (skip)
										newPath.append(sp);
									skip = true;
								}
				return SearchEntryPath(dynamic_cast<DatFileEntryDirectory*>(de), newPath);
			}
		}
	}

	return nullptr;
}

/**
 * Get a list of files corresponding to the specified pattern.
 *
 * @param virtualDirPath	Origin of browsing
 * @param pattern			Search filename pattern
 * @param recursive			Do the search inside subfolders
 *
 * @return List of references to the files corresponding to pattern
 */
vector<DatFileEntry*> Erso::getElements( path virtualDirPath, string pattern, bool recursive ) {
	vector<DatFileEntry*> result;
	DatFileEntryDirectory* curDir = dynamic_cast<DatFileEntryDirectory*>(getElement(virtualDirPath));

	if (recursive) {
		auto recursFileList = SearchEntriesPattern(curDir, pattern);
		result.insert(result.end(), recursFileList.begin(), recursFileList.end());
	} else {
		for ( DatFileEntry* e : curDir->getFiles() ) {
			if (e->getName().find(pattern) != string::npos)	//TODO: Replace using RegEx search
				result.push_back(e);
		}
	}

	return result;
}

/**
 * Browse inside virtual path to the final object from specific root DatFileEntryDirectory.
 *
 * @param curDir			Origin of browsing
 * @param virtualDirPath	Requested file path
 *
 * @return Reference to the file resource
 */
DatFileEntry* Erso::SearchEntryPath( DatFileEntryDirectory* curDir, path virtualDirPath ) {
	bool skip = false;
	path newPath;

	for ( DatFileEntry* e : curDir->getFiles() ) {
		if (virtualDirPath.size() == 1 || virtualDirPath.begin() == --virtualDirPath.end()) {
			if (e->getName() == virtualDirPath.filename().string())
				return e;
		} else {
			if (e->isDirectory() && (e->getName() == virtualDirPath.begin()->string())) {
				for (auto& sp : virtualDirPath) {
					if (skip)
						newPath.append(sp);
					skip = true;
				}

				return SearchEntryPath(dynamic_cast<DatFileEntryDirectory*>(e), newPath);
			}
		}
	}

	return nullptr;
}

/**
 * Browse inside virtual path to find specific pattern and add it reference to the list in input.
 *
 * @param upDir			Origin of browsing
 * @param pattern		Search filename pattern
 *
 * @return				Vector list of DatFileEntry
 */
vector<DatFileEntry*> Erso::SearchEntriesPattern( DatFileEntryDirectory* upDir, string pattern ) {
	vector<DatFileEntry*> subDatEntries;

	for ( DatFileEntry* e : upDir->getFiles() )
		if (e->getName().find(pattern) != string::npos)	//TODO: Replace using RegEx search
			subDatEntries.push_back(e);

	for ( DatFileEntry* e : upDir->getFiles() ) {
		if (e->isDirectory()) {
			auto recursFileList = SearchEntriesPattern(dynamic_cast<DatFileEntryDirectory*>(e), pattern);
			subDatEntries.insert(subDatEntries.end(), recursFileList.begin(), recursFileList.end());
		}
	}

	return subDatEntries;
}

}
