/**
 * @file		Erso.hpp
 * @date		04/02/2023
 * @author		JackCarterSmith
 * @copyright	GPL-v3.0
 * @brief		Rogue DATA/BUNDLE file parser and interface.
 * @warning		This module should be initialized before every others modules call of RDI library.
 *
 * Erso is the first module of the RDI library, it's use research work of
 * <rogue_dev_team>, dpethes and myself to parse legacy data files of Rogue
 * game.
 * Currently compatible legacy files are :
 * - DATA.DAT/HDR: as primary elements package containing all levels, sound,
 * 		textures, etc.
 * - BUNDLE.000: as title menu resources like pictures, buttons and cursor
 * 		textures, menu sounds, menu macro-scripting.
 * - BUNDLE.001: not known for now.
 *
 * This module should be initialized before every others modules call of RDI
 * library.
 *
 * Standard usage:
 * - Create a single Erso object with the folder path of datas files as input
 * argument.
 * - All supported data files will be parsed and a virtual tree view will be
 * created.
 * - The virtual tree view can be browse using appropriate 'get' functions.
 *
 */

#include <boost/filesystem.hpp>
#include <FileHandler/Generic.hpp>
#include <vector>
#include <RDI.hpp>	//TODO: Should be removed with RDI_RESULT

#include "DatFileEntry.hpp"


#ifndef ERSO_HPP_
#define ERSO_HPP_

namespace RDI {

struct dataSection {
	std::string name = "";
	unsigned int offset = 0;
};

/**
 * Erso is class codename of the DATA/BUNDLE file parser of RDI library.
 *
 * @pre RDI should construct Erso class before doing anything on datas.
 */
class Erso final {
public:
	Erso( std::string fPath );
	Erso( std::string fPath, RDI_RESULT* outResult );
	~Erso();

	MEMFILE getRDat() const { return rDatPtr; }

	/* Workspace management */
	std::string getWorkingDirectory() const { return workingDir; }
	RDI_RESULT setWorkingDirectory( std::string newPath );

	/* Generals informations getters */
	unsigned char getDataSectionCount() const { return cDatSection; }
	std::string getDataSectionName( unsigned char id ) const;
	unsigned int getDataSectionOffset( unsigned char id ) const;
	bool isUpdatedVersion() const { return (cDatSection > 2); }

	/* Tree manipulation/parse methods */
	DatFile::DatFileEntry* getElement( boost::filesystem::path virtualDirPath );
	std::vector<DatFile::DatFileEntry*> getElements( boost::filesystem::path virtualDirPath, std::string pattern, bool recursive );

private:
	std::string workingDir = ".";

	unsigned char cDatSection = 0;
	std::vector<struct dataSection> *pDatSection = nullptr;
	MEMFILE rDatPtr = nullptr;

	std::vector<DatFile::DatFileEntryDirectory*> *pFTRoot = nullptr;


	/* File processing methods */
	RDI_RESULT Process();
	RDI_RESULT DataFilesDumping();
	MEMFILE CreateMemFile( boost::filesystem::path filePath );
	RDI_RESULT CreateDataFilesTree();
	RDI_RESULT ProcessTreeDirectoryContents(
			DatFile::DatFileEntryDirectory* pCurDir,
			MEMFILE pDesc,
			const unsigned int count,
			MEMFILE pSectionStart
		);
	void DisposeFile();

	/* Tree manipulation/parse methods */
	DatFile::DatFileEntry* SearchEntryPath( DatFile::DatFileEntryDirectory *curDir, boost::filesystem::path virtualDirPath);
	std::vector<DatFile::DatFileEntry*> SearchEntriesPattern(DatFile::DatFileEntryDirectory *upDir, std::string pattern);

};

} /* namespace RDI */

#endif /* ERSO_HPP_ */
