/**
 * @file		Krennic.cpp
 * @date		18/09/2023
 * @author		JackCarterSmith
 * @copyright	GPL-v3.0
 * @brief		Main game assets parser and interface.
 *
 * Complete class would take Erso legacy files tree and extract levels, objects,
 * textures, animations, etc. adapted datas for OpenGL/Direct3D application.
 *
 */

#include <stdexcept>
#include <vector>
#include <array>
#include <RSPModel.h>
#include <RSPTerrain.h>
#include <RSPTexture.h>
#include <boost/filesystem.hpp>
#include "Erso.hpp"
#include "Krennic.hpp"

#include <FileHandler/Generic.hpp>
#include <FileHandler/HMT.hpp>
#include <FileHandler/HOB.hpp>

using std::string;
using std::vector;
using std::array;

using namespace RDI::DatFile;


namespace RDI {

struct PathDescriptor {
	PathDescriptor(string p, bool r) noexcept
		: path(p), isRecursive(r) {}

	PathDescriptor(const PathDescriptor&) = default;
	PathDescriptor& operator=(const PathDescriptor&) = default;

	PathDescriptor(PathDescriptor&&) = default;
	PathDescriptor& operator=(PathDescriptor&&) = default;

	string	path;
	bool	isRecursive;
};

Krennic::Krennic( Erso* pErso ) {
	BuildLevelFileList(pErso);
	BuildModelFileList(pErso);
	BuildTextureFileList(pErso);
	BuildMusicFileList(pErso);
	BuildSampleFileList(pErso);
}

Krennic::~Krennic() {}


void Krennic::BuildLevelFileList( Erso* pErso ) {
	const static array<PathDescriptor, 1> legacyLvlPath = {
		PathDescriptor("data/level", true)
	};

	DatFileEntryFile* datFile = nullptr;
	DatFileEntryFile* hmpFile = nullptr;
	DatFileEntryFile* gvmFile = nullptr;

	listLevelsName.clear();
	for (PathDescriptor legacyLvlSubpath : legacyLvlPath) {
		for (DatFileEntry* file : pErso->getElements(boost::filesystem::path(legacyLvlSubpath.path), "lv_", legacyLvlSubpath.isRecursive)) {
			if (file->isDirectory()) {
				DatFileEntryDirectory* lvlDir = dynamic_cast<DatFileEntryDirectory*>(file);
				for ( DatFileEntry* lvlFile : lvlDir->getFiles() ) {
					if (lvlFile->getName().compare("dat") == 0)
						datFile = dynamic_cast<DatFileEntryFile*>(lvlFile);
					else if (lvlFile->getName().compare("hmp") == 0)
						hmpFile = dynamic_cast<DatFileEntryFile*>(lvlFile);
					else if (lvlFile->getName().compare("gvm") == 0)
						gvmFile = dynamic_cast<DatFileEntryFile*>(lvlFile);
				}

				if (datFile == nullptr || hmpFile == nullptr || gvmFile == nullptr)
					throw std::runtime_error("One or more level files can't be found!");

				auto lvlObj = new LVL(datFile, hmpFile, gvmFile);
				listLevels.push_back(lvlObj);
				listLevelsName.push_back(lvlObj->getFullPath());
			}
		}
	}
}

void Krennic::BuildModelFileList( Erso* pErso ) {
	const static array<PathDescriptor, 9> legacyModelPath = {
		PathDescriptor("data", false),
		PathDescriptor("data/pl_crafts", false),
		PathDescriptor("data/reb_stuff", false),
		PathDescriptor("data/imp_stuff", false),
		PathDescriptor("data/gnrc_stuff", false),
		PathDescriptor("data/level", true),
		PathDescriptor("data/frontend", true),
		PathDescriptor("data/dbg", false),
		PathDescriptor("dbg_data", false)
	};
	// Update version game legacy assets paths
	const static array<PathDescriptor, 1> legacyModelPath2= {
		PathDescriptor("data2", false)
	};

	listModels.clear();
	listModelsName.clear();
	for (PathDescriptor legacyModelSubpath : legacyModelPath) {
		for (DatFileEntry* file : pErso->getElements(boost::filesystem::path(legacyModelSubpath.path), "_HOB", legacyModelSubpath.isRecursive)) {
			auto hobObj = new HOB(dynamic_cast<DatFileEntryFile*>(file));
			listModels.push_back(hobObj);
			listModelsName.push_back(hobObj->getFullPath());
		}
	}
	if (pErso->isUpdatedVersion()) {
		for (PathDescriptor legacyModelSubpath2 : legacyModelPath2) {
			for (DatFileEntry* file : pErso->getElements(boost::filesystem::path(legacyModelSubpath2.path), "_HOB", legacyModelSubpath2.isRecursive)) {
				auto hobObj = new HOB(dynamic_cast<DatFileEntryFile*>(file));
				listModels.push_back(hobObj);
				listModelsName.push_back(hobObj->getFullPath());
			}
		}
	}
}

void Krennic::BuildTextureFileList( Erso* pErso ) {
	// Base game legacy assets paths
	const static array<PathDescriptor, 9> legacyTexturePath = {
		PathDescriptor("data", false),
		PathDescriptor("data/pl_crafts", false),
		PathDescriptor("data/reb_stuff", false),
		PathDescriptor("data/imp_stuff", false),
		PathDescriptor("data/gnrc_stuff", false),
		PathDescriptor("data/level", true),
		PathDescriptor("data/frontend", true),
		PathDescriptor("data/dbg", false),
		PathDescriptor("dbg_data", false)
	};
	// Update version game legacy assets paths
	const static array<PathDescriptor, 1> legacyTexturePath2= {
		PathDescriptor("data2", false)
	};

	listTexturesName.clear();
	for (PathDescriptor legacyTextureSubpath : legacyTexturePath) {
		for (DatFileEntry* file : pErso->getElements(boost::filesystem::path(legacyTextureSubpath.path), "_HMT", legacyTextureSubpath.isRecursive)) {
			//TODO: texture-class builder
			listTexturesName.push_back(file->getPath());
		}
	}
	if (pErso->isUpdatedVersion()) {
		for (PathDescriptor legacyTextureSubpath2 : legacyTexturePath2) {
			for (DatFileEntry* file : pErso->getElements(boost::filesystem::path(legacyTextureSubpath2.path), "_HMT", legacyTextureSubpath2.isRecursive)) {
				//TODO: texture-class builder
				listTexturesName.push_back(file->getPath());
			}
		}
	}
}

void Krennic::BuildMusicFileList( Erso* pErso ) {
	const static array<PathDescriptor, 1> legacyLvlPath = {
		PathDescriptor("data/sound", false)
	};

	listMusicsName.clear();
	for (PathDescriptor legacyMusicSubpath : legacyLvlPath) {
		for (DatFileEntry* file : pErso->getElements(boost::filesystem::path(legacyMusicSubpath.path), "_SNG", legacyMusicSubpath.isRecursive)) {
			//TODO: MusyX-Class builder
			listMusicsName.push_back(file->getPath());
		}
	}
}

void Krennic::BuildSampleFileList( Erso* pErso ) {

}

LVL* Krennic::getLevel( string name ) {
	unsigned int i = 0;

	for (string fileName : listLevelsName) {
		if (fileName.find(name) != string::npos) {
			return listLevels.at(i);
		}
		i++;
	}

	throw std::invalid_argument("Krennic can't found the requested level.");
}

HOB* Krennic::getModel( string name ) {
	unsigned int i = 0;

	for (string fileName : listModelsName) {
		if (fileName.find(name) != string::npos) {
			return listModels.at(i);
		}
		i++;
	}

	throw std::invalid_argument("Krennic can't found the requested model.");
}

//HMT* Krennic::getTexture( string name ) {}

//void* Krennic::getMusic( string name ) {}

//void* Krennic::getSample( string name ) {}


void Krennic::DisposeLevels() {}

void Krennic::DisposeModels() {
	for (HOB* hobObj : listModels) {
		if (hobObj->isLoaded())
			hobObj->Dispose();

		delete hobObj;
	}
}

void Krennic::DisposeTextures() {}

void Krennic::DisposeMusics() {}

void Krennic::DisposeSamples() {}

}
