/**
 * @file		KrennicHandlerHOB.cpp
 * @date		19/01/2023
 * @author		JackCarterSmith
 * @copyright	GPL-v3.0
 * @brief		HOB file object class.
 *
 */

#include <string>
#include <stdexcept>
#include <boost/qvm_lite.hpp>
#include <FileHandler/Generic.hpp>
#include <RSPModel.h>
#include <RSPModel_errordefs.h>
#include "DatFileEntry.hpp"
#include <FileHandler/HOB.hpp>

using std::runtime_error;
using std::out_of_range;
using std::invalid_argument;
using std::string;
using namespace RDI::DatFile;
using namespace boost::qvm;


namespace RDI {

HOB::Mesh* BuildObjectMesh( const T_RSPMODEL_OBJECT* hobPtr );
HOB::Mesh* FillMesh( const T_RSPMODEL_OBJ_PARTS* objPtr );


/* -------------------------------------------------------------------------- */
HOB::HOB( DatFileEntryFile* hDat )
		: GenericFile(hDat) {
	this->fileExtension = "hob";
}

HOB::~HOB() {
	Dispose();
}


/* -------------------------------------------------------------------------- */
void HOB::Load() {
	unsigned int i;
	RSPMODEL_PARAMETERS rspParams;
	unsigned short result = RSPLIB_SUCCESS;

	if (this->isLoaded()) return;

	this->hobInstance = new T_RSPMODEL_HOB{};
	auto hobInstancePtr = static_cast<T_RSPMODEL_HOB*>(this->hobInstance);
	rspParams.raw = 0;
	result = RSPModel_processHOBFileMemory(hobInstancePtr, this->pMemLoc, this->size, rspParams);
	if (result != RSPLIB_SUCCESS) {
		RSPModel_freeHOB(hobInstancePtr);
		delete static_cast<T_RSPMODEL_HOB*>(this->hobInstance);

		if (result == RSPLIB_ERROR_GENERIC)
			throw out_of_range("HOB handler tried to process empty file.");
		else
			throw runtime_error("HOB handler processing failed with error: " + std::to_string(result));
	}

	this->objCnt = hobInstancePtr->obj_count;
	this->objectMeshList.clear();
	for ( i = 0; i < this->objCnt; i++ )
		this->objectMeshList[string(hobInstancePtr->objects[i].name)] = BuildObjectMesh(hobInstancePtr->objects + i);

	RSPModel_freeHOB(hobInstancePtr);

	// Must be called last to set loaded state
	GenericFile::Load();
}

void HOB::Dispose() {
	if (!this->isLoaded()) return;

	for ( const auto& [objName, objMesh] : this->objectMeshList ) {
		for ( auto v : objMesh->vertices)
			delete v;

		for ( auto sm : objMesh->subMeshs)
			delete sm;

		delete objMesh;
	}

	this->objectMeshList.clear();
	delete static_cast<T_RSPMODEL_HOB*>(this->hobInstance);

	// Must be called last to unset loaded state
	GenericFile::Dispose();
}


/* -------------------------------------------------------------------------- */
unsigned int HOB::getObjectCount() const {
	if (this->isLoaded())
		return this->objCnt;

	return RSPModel_getHOBFileMemObjCount(this->pMemLoc);
}

std::string* HOB::getObjectName( unsigned short index ) const {
	if (!this->isLoaded())
		throw runtime_error("HOB handler has not been loaded prior to function calling.");

	if (index >= this->objCnt)
		throw out_of_range("HOB handler object index out of range!");

	auto entry = objectMeshList.begin();
	std::advance(entry, index);
	return new string(entry->first);
}

HOB::Mesh* HOB::getMeshFromObject( unsigned short index ) const {
	if (!this->isLoaded())
		throw runtime_error("HOB handler has not been loaded prior to function calling.");

	if (index >= this->objCnt)
		throw out_of_range("HOB handler object index out of range!");

	auto entry = objectMeshList.begin();
	std::advance(entry, index);
	return entry->second;
}

HOB::Mesh* HOB::getMeshFromObject( string name ) const {
	if (!this->isLoaded())
		throw runtime_error("HOB handler has not been loaded prior to function calling.");

	for ( const auto& [objName, objMesh] : this->objectMeshList )
		if (objName == name) return objMesh;

	throw invalid_argument("HOB handler can't find requested object name.");
}


/* -------------------------------------------------------------------------- */
HOB::Mesh* BuildObjectMesh( const T_RSPMODEL_OBJECT* hobPtr ) {
	unsigned int i;
	auto meshObj = new HOB::Mesh();

	meshObj->indices.clear();
	meshObj->vertices.clear();
	meshObj->subMeshs.clear();

	meshObj->transform = identity_mat<float, 4>();
	for (i = 0; i < hobPtr->object_part_count; i++)
		meshObj->subMeshs.push_back(FillMesh(hobPtr->object_parts + i));

	return meshObj;
}

HOB::Mesh* FillMesh( const T_RSPMODEL_OBJ_PARTS* objPtr ) {
	unsigned int i;
	auto meshObj = new HOB::Mesh();

	meshObj->indices.clear();
	meshObj->vertices.clear();
	meshObj->subMeshs.clear();

	//meshObj->transform = mat<float,4,4>();
	meshObj->transform = identity_mat<float, 4>();
	for (i = 0; i < objPtr->vertex_count; i++) {
		auto v = new vec<float, 3>();
		v->a[0] = static_cast<float>(-(objPtr->vertices[i].x)) / 1024.0;
		v->a[1] = static_cast<float>(-(objPtr->vertices[i].y)) / 1024.0;
		v->a[2] = static_cast<float>((objPtr->vertices[i].z)) / 1024.0;
		meshObj->vertices.push_back(v);
	}
	for (i = 0; i < objPtr->face_count; i++) {
		meshObj->indices.push_back(static_cast<uint16_t>(objPtr->faces[i].indices[0]));
		meshObj->indices.push_back(static_cast<uint16_t>(objPtr->faces[i].indices[1]));
		meshObj->indices.push_back(static_cast<uint16_t>(objPtr->faces[i].indices[2]));

		if (objPtr->faces[i].flags_bits.fIsQuad) {
			meshObj->indices.push_back(static_cast<uint16_t>(objPtr->faces[i].indices[2]));
			meshObj->indices.push_back(static_cast<uint16_t>(objPtr->faces[i].indices[3]));
			meshObj->indices.push_back(static_cast<uint16_t>(objPtr->faces[i].indices[0]));
		}
	}

	return meshObj;
}

}
