/**
 * @file		KrennicHandlerLVL.cpp
 * @date		07/02/2023
 * @author		JackCarterSmith
 * @copyright	GPL-v3.0
 * @brief		LVL object class.
 *
 */

#include <string>
#include <stdexcept>
#include <boost/filesystem.hpp>
#include <boost/qvm_lite.hpp>
#include <FileHandler/Generic.hpp>
#include <RSPTerrain.h>
#include <RSPTerrain_errordefs.h>
#include "DatFileEntry.hpp"
#include <FileHandler/LVL.hpp>

using std::runtime_error;
using std::invalid_argument;
using std::string;
using namespace RDI::DatFile;
using namespace boost::qvm;


namespace RDI {

LVL::Mesh* BuildMapMesh( const T_RSPTERRAIN_MESH& hobPtr );


/* -------------------------------------------------------------------------- */
LVL::LVL( DatFileEntryFile* hDat, DatFileEntryFile* hHmp, DatFileEntryFile* hGvm )
		: GenericFile(hDat), hmpFileSize(hHmp->getSize()), gvmFileSize(hGvm->getSize()),
		  pHmpMemLoc(hHmp->getDatas()), pGvmMemLoc(hGvm->getDatas()) {
	this->fileExtension = "lvl";
	this->levelName = boost::filesystem::path(hDat->getPath()).parent_path().filename().string();
	this->fullLevelDirPath = boost::filesystem::path(hDat->getPath()).parent_path().string();
}

LVL::~LVL() {
	Dispose();
}


/* -------------------------------------------------------------------------- */
void LVL::Load() {
	RSPTERRAIN_PARAMETERS rspParams;
	T_RSPTERRAIN_MESH intermediateMesh;
	unsigned short result = RSPLIB_SUCCESS;

	if (this->isLoaded()) return;

	this->hmpInstance = new T_RSPTERRAIN_HMP{};
	auto hmpInstancePtr = static_cast<T_RSPTERRAIN_HMP*>(this->hmpInstance);
	rspParams.raw = 0;
	rspParams.invertZ = true;

	result = RSPTerrain_processHMPFileMemory(hmpInstancePtr, this->pHmpMemLoc, this->hmpFileSize, rspParams);
	if (result != RSPLIB_SUCCESS) {
		RSPTerrain_freeHMP(hmpInstancePtr);
		delete static_cast<T_RSPTERRAIN_HMP*>(this->hmpInstance);

		throw runtime_error("HMP handler processing failed with error: " + std::to_string(result));
	}

	// Use RSPTerrainLib to build our level mesh and convert it to generic mesh format
	RSPTerrain_terrainToMesh(&intermediateMesh, hmpInstancePtr, 0.1f);
	this->mapMesh = BuildMapMesh(intermediateMesh);

	RSPTerrain_freeHMP(hmpInstancePtr);

	// Must be called last to set loaded state
	GenericFile::Load();
}

void LVL::Dispose() {
	if (!this->isLoaded()) return;

	delete this->mapMesh;
	delete static_cast<T_RSPTERRAIN_HMP*>(this->hmpInstance);

	// Must be called last to unset loaded state
	GenericFile::Dispose();
}


/* -------------------------------------------------------------------------- */



/* -------------------------------------------------------------------------- */
LVL::Mesh* BuildMapMesh( const T_RSPTERRAIN_MESH& terrainMesh ) {
	unsigned int i,j;
	auto mapMesh = new LVL::Mesh();

	mapMesh->indices.clear();
	mapMesh->vertices.clear();
	mapMesh->indices.reserve((terrainMesh.height-1) * (terrainMesh.width-1));
	mapMesh->vertices.reserve(terrainMesh.vertices_count);

	for (i = 0; i < terrainMesh.vertices_count; i++) {
		auto v = new vec<float, 3>();

		v->a[0] = terrainMesh.verticesmap[i].x;
		v->a[1] = terrainMesh.verticesmap[i].y;
		v->a[2] = terrainMesh.verticesmap[i].z;

		mapMesh->vertices.push_back(v);
	}

	for ( i = 0; i < static_cast<unsigned int>(terrainMesh.height) - 1; i++ ) {
		for ( j = 0; j < static_cast<unsigned int>(terrainMesh.width) - 1; j++ ) {
			//uint16_t indice = static_cast<uint16_t>(i * terrainMesh.width + j);
			uint32_t indice = static_cast<uint32_t>(i * terrainMesh.width + j);

			mapMesh->indices.push_back(indice);
			mapMesh->indices.push_back(indice + 1);
			mapMesh->indices.push_back(indice + terrainMesh.width);

			mapMesh->indices.push_back(indice + 1);
			mapMesh->indices.push_back(indice + terrainMesh.width + 1);
			mapMesh->indices.push_back(indice + terrainMesh.width);
		}
	}

	return mapMesh;
}

}
